C   eigenplane - find basis vectors for eigenplanes of an isometry of S^3 represented as a 4x4 matrix
C
C   Copyright (C) 2004 Boud Roukema
C
C   This program is free software; you can redistribute it and/or modify
C   it under the terms of the GNU General Public License as published by
C   the Free Software Foundation; either version 2, or (at your option)
C   any later version.
C
C   This program is distributed in the hope that it will be useful,
C   but WITHOUT ANY WARRANTY; without even the implied warranty of
C   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C   GNU General Public License for more details.
C
C   You should have received a copy of the GNU General Public License
C   along with this program; if not, write to the Free Software Foundation,
C   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

c     --- this is the main fortran subroutine for the eigenplane program,
c     following the C formal main program and the mainsub fortran 
c     command line check --

      subroutine eigmain

      include 'eigcom.f'

c      real*8 gen(4,4) ! 4x4 matrix =  isometry and maybe = generator
c      real*8 gtemp(4,4) ! 4x4 matrix  - temporary working matrix
c      real*8 rbasis(4,4) ! a random orthonormal basis
c      real*8 rbinv(4,4)        ! a random orthonormal basis
      real*8 genbas(4,4) ! the generator in the basis rbasis

      real*8 uguess(4) ! initial guess at a vector in an eigenplane
      real*8 soln(4,4) ! four columns (.,i) = solution vectors in eigenplanes

c      real*8 chi, ang_u, det

c      real*8 theta,dtheta, phi,dphi ! angles and precision

      character*1  ch

c      logical*4    is_input ! SHIFT TIHS TO A COMMAND LINE PARAMETER! NOW!
c      data is_input /.false./

      data ch /'n'/

c     testing Makefile dependence on included file
c      print*,'eigmain: toldef=',toldef 

c     --- flush generator ---
      do i=1,iseed 
         x=randt()
c     print*,'x=',x
      enddo

c      if(debug)then
c         read(5,*)x
c         print*,'read x=',x
c         print*,' '
c      endif


      if(.not.(q_random))then
         if(q_input)then
            open(iifile,file=inamc,access='sequential',
     _           status='old')
         endif

         read(iifile,*)genbas

         if(q_input)then
            close(iifile)
         endif

         if(q_verb)call printmat(genbas)
      else
c     -- Generate a random isometry of S^3 in R^4
c     -- expressed in a randomly oriented orthonormal basis
         call getiso(genbas,th_in,ph_in) ! 3 outputs
      endif


c     ---- start with an arbitrary vector ----
      do i=1,4
         uguess(i)=0.0d0
      enddo
      uguess(1)=1.0d0

      call getang(genbas,uguess,thet_s,soln(1,1))! 1 input, 2 outputs

c     --- 2nd iteration to be sure 
      call getang(genbas,soln(1,1),thet_s,soln(1,1))! 1 input, 2 outputs


      call getfour(genbas,soln,thet2,phi_s)

      if(.not.(q_warn))then
         if(abs(thet_s-thet2).gt.toldef*10.0d0)print*,'WARNING:',
     _        ' thet_s,thet2=',thet_s,thet2
      endif




      if(debug)then
         print*,'q_output,iofile,onamc=',
     _        q_output,iofile,onamc
      endif
      if(q_output)then
         open(iofile,file=onamc,access='sequential',status='unknown')
      endif
         

      if(q_random)then
         print*,'The input angles of the random isometry'
         print*,'were th_in, ph_in=', th_in,ph_in
         print*,'The angles and eigenplane vectors'
         print*,'found by this algorithm are:'
      endif
      write(iofile,'(2f14.6)')thet_s,phi_s


c     --- output written so that text file visually has vectors as columns ---
      do i=1,4
         do j=1,4
            write(iofile,'(f14.6,$)')soln(i,j) ! write by rows
         enddo
         write(iofile,'(a)')' '
      enddo
      if(q_output)close(iofile)



      return
      end



c     ==== chi is a parameter which should be zero iff 
c     the vector  u  is in an eigenplane of  M 
c     It is not really a chi^2 parameter,  chi is just a convenient
c     name to remember for the parameter.
c     ==== 

      subroutine getchi(u,gen, chi, angle)

      include 'eigcom.f'

c     INPUTS:
      real*8 u(4)  ! vector
      real*8 gen(4,4) ! 4x4 matrix =  isometry and maybe = generator

C     OUTPUTS:
      real*8 chi                ! nearness to eigenplane parameter
      real*8 angle              ! the rotation angle

C     INTERNAL:
      real*8 v(4),v2(4),a1,a2, d


C     -- first line from blas dgemv --
c            SUBROUTINE DGEMV ( TRANS, M, N, ALPHA, A, LDA, X, INCX,
c     $                   BETA, Y, INCY )
      
c     ---  v  = GEN * u --- 
      do i=1,4
         v(i)=zerovec(i)
      enddo
      
      call dgemv('n', 4,4, 1.0d0, gen, 4, u,
     _     1, 0.0d0, v, 1)

c     ---  d = u.v = u.(GEN u)  dot  product ---
      d = ddot(4, u, 1, v, 1) 

      a1 = acos(max(min(d,1.0d0-toldef),-1.0d0+toldef)) ! angle 1

c     This angle should be less than pi, since it should be an integer
c     fraction of  2pi.


c     ---  v2  = GEN * v = GEN * GEN * u --- 
      do i=1,4
         v2(i)=zerovec(i)
      enddo
      call dgemv('n', 4,4, 1.0d0, gen, 4, v,
     _     1, 0.0d0, v2, 1)

c     ---  d = u.v2 = u.(GEN GEN u)  dot  product ---
      d = ddot(4, u, 1, v2, 1) 

      a2 = acos(max(min(d,1.0d0-toldef),-1.0d0+toldef)) ! angle 2

c     This angle might be more than pi, since it is (if u is in an 
c     eigenplane) twice the angle  a1 , i.e. it could be  (2*pi-a2).
c     How do we decide which is the case? 
c     
c     We don't, testing the wrong cos^-1 solution will not erroneously
c     give an angle which matches with 2 * a1. So it is enough to
c     test both possibilities and take the minimum.

      b1= 2.0d0*a1 - a2
      b2= 2.0d0*a1 - (2.0d0*pi - a2)


      chi = min( b1*b1, b2*b2 )
      angle = a1

      return
      end


c     === rotate matrix  g_in  by theta in the i1-i2 plane ===
      subroutine ranrot(i1,i2,theta,g_in,g_out)
      include 'eigcom.f'

C     INPUTS:
      integer i1,i2 
      real*8  theta
      real*8  g_in(4,4)

C     OUTPUTS:
      real*8  g_out(4,4)

C     INTERNAL:
      real*8  dummy(4,4)
      real*8  rot(4,4)
      character*1  ch

      data ch /'n'/

      if(i1.lt.1.or.i2.lt.1.or.i1.gt.4.or.i2.gt.4
     _     .or.i1.eq.i2)then
         print*,'ERROR: i1,i2 = ',i1,i2
         stop
      endif

      do j=1,4
         do i=1,4
            rot(i,j)=0.0d0
            dummy(i,j)=0.0d0
            if(i.eq.j) rot(i,j)=1.0d0
         enddo
      enddo
      rot(i1,i1)=cos(theta)
      rot(i1,i2)=-sin(theta)
      rot(i2,i1)=sin(theta)
      rot(i2,i2)=cos(theta)
            
      call dgemm(ch,ch, 4,4,4, 1.0d0, g_in, 4, rot, 4, 
     _     1.0d0, dummy, 4)

      do j=1,4
         do i=1,4
            g_out(i,j)=dummy(i,j)
         enddo
      enddo
      return
      end


c     ==== calculate a rotation matrix in the i1-i2 direction by angle theta =====
      subroutine idrot(i1,i2,theta,g_out)
      include 'eigcom.f'

C     INPUTS:
      integer i1,i2 
      real*8  theta

C     OUTPUTS:
      real*8  g_out(4,4)

c     internal:
      real*8  g_in(4,4)

      do j=1,4
         do i=1,4
            if(i.eq.j)then
c               g_in(i,j)=1.0d0
               g_out(i,j)=1.0d0
            else
c               g_in(i,j)=0.0d0
               g_out(i,j)=0.0d0
            endif
         enddo
      enddo
      
c      call ranrot(i1,i2,theta,g_in,g_out)
      g_out(i1,i1)= cos(theta)
      g_out(i1,i2)= -sin(theta)
      g_out(i2,i1)= sin(theta)
      g_out(i2,i2)= cos(theta)


c     -- g_out from ranrot is g_out from this routine --

      return
      end



c     ==== print a matrix  gen  to stdout ====
      subroutine printmat(gen)
      include 'eigcom.f'

c     INPUT:
      real*8   gen(4,4)
      

      if(q_verb)print*,'The matrix is: '
c     --- note: the i,j ordering for printed output is along rows not columns ---
      do i=1,4
         do j=1,4
            write(6,'(f11.4,$)')gen(i,j)
         enddo
         print*,' '
      enddo
      print*,' '

      return
      end


c     --- This calls lapack to calculate a determinant, but
c     is not really needed, so to reduce unnecessary
c     compilation dependencies, this is commented out.

Cc     ==== find the determinant of a 4x4 matrix using lapack ====
C      real*8  function det4x4(gen)
C
C      include 'eigcom.f'
C
Cc     INPUTS:
C      real*8 gen(4,4) ! 4x4 matrix =  isometry and maybe = generator
C
Cc     INTERNAL:
C      real*8 gen_lu(4,4) ! 4x4 matrix =  isometry and maybe = generator
C      integer  ipiv(4),ipiv_2(4)
C
C      do j=1,4
C         do i=1,4
C            gen_lu(i,j)=gen(i,j) ! initialise gen_lu
Cc     if(q_verb)print*,'i,j,gen(i,j)=',i,j,gen(i,j)
C         enddo
C      enddo
C
C      if(q_verb)print*,'Find its determinant using lapack and',
C     _     ' normalise it...'
C
C      call dgetrf(4,4, gen_lu, 4, ipiv, info)
C      if(q_verb)print*,'dgetrf: info=',info
C
C      det= gen_lu(1,1)*gen_lu(2,2)*gen_lu(3,3)*gen_lu(4,4)
C
C
Cc      if(q_verb)print*,'product of LU diagonal is: ',det
Cc     -- find determinant +1 or -1 of ipiv(4) --
C      do i=1,4
C         ipiv_2(i)=ipiv(i)
C      enddo
C
Cc     --- 1st row
C      idet=1
C      do i=2,ipiv_2(1)
C         idet=idet* (-1) ! 1st row
C      enddo
C      do i=2,4
Cc     -- remove  ipiv_2(1)-th column from matrix 
C         if(ipiv_2(i).gt.ipiv_2(1))ipiv_2(i)=ipiv_2(i)-1
C      enddo
C
Cc     --- 2nd row
C      idet=1
C      do i=2,ipiv_2(2)
C         idet=idet*(-1) ! 2nd row
C      enddo
C      do i=2,3
Cc     -- remove  ipiv_2(2)-th column from matrix 
C         if(ipiv_2(i).gt.ipiv_2(2))ipiv_2(i)=ipiv_2(i)-1
C      enddo
C
Cc     --- 3rd row : only two possibilities left
C      idet=1
C      if(ipiv_2(3).eq.2) idet=idet*(-1)
C
C      det=det*dble(idet)
C      if(abs(det).lt.toldef)then
C         print*,'det=',det
C         stop 'matrix is singular - try --seed=<some integer>'
C      endif
C
C      if(q_verb)print*,'det=',det
C
C      det4x4 = det
C      return
C      end


c     ==== generate a random isometry (not holonomy) of S^3 in R^4 ====
      subroutine r_isom(gen,theta,phi)

      include 'eigcom.f'

C     OUTPUTS: 
      real*8  gen(4,4)  ! input matrix
      real*8  theta,phi ! rotation angles
C     INTERNAL:
      real*8  gtemp(4,4)


c     --- identity matrix ---
      do j=1,4
         do i=1,4
            gen(i,j)=0.0d0
            if(i.eq.j)gen(i,j)=1.0d0
         enddo
      enddo

      if(q_verb)then
         print*,'identity: '
         call printmat(gen)
      endif

c      det=det4x4(gen)
c      if(q_verb)print*,"Its determinant is",det

      theta=randt()*pi
      i1=nint( randt()*3.98 +0.51 )
      i2=nint( randt()*2.98 +0.51 )
      if(i1.eq.i2)i2=i2+1
      if(q_verb)print*,'i1,i2,theta=',i1,i2,theta
      call ranrot(i1,i2, theta, gen, gtemp)
      do j=1,4
         do i=1,4
            gen(i,j)=gtemp(i,j)
         enddo
      enddo

      if(q_verb)then
         print*,'after rotation by theta: '
         call printmat(gen)
      endif

      phi=randt()*pi
      i=1
      do while((i.eq.i1.or.i.eq.i2).and.i.lt.4)
         i=i+1
      enddo
      j1=i
      i=1
      do while((i.eq.i1.or.i.eq.i2.or.i.eq.j1).and.i.lt.4)
         i=i+1
      enddo
      j2=i
      if(q_verb)print*,'j1,j2,phi=',j1,j2,phi
      call ranrot(j1,j2, phi, gen, gtemp)
      do j=1,4
         do i=1,4
            gen(i,j)=gtemp(i,j)
         enddo
      enddo

      if(q_verb)then
         print*,'after rotation by phi: '
         call printmat(gen)
      endif

      return
      end


      subroutine dummy1  ! scratch pad only
      include 'eigcom.f'

C     OUTPUTS: 
      real*8  gen(4,4)

c     ---- a generally random matrix, not very useful
      do j=1,4
         do i=1,4
            gen(i,j)=2.0*randt()-1.0d0
            if(q_verb)print*,'i,j,gen(i,j)=',i,j,gen(i,j)
         enddo
      enddo

      if(q_verb)print*,'The matrix is: '
      do j=1,4
         do i=1,4
            gen(i,j)=gen(i,j)/det
            if(q_verb)write(6,'(f9.2,$)')gen(i,j)
         enddo
         if(q_verb)print*,' '
      enddo
      if(q_verb)print*,' '


      return
      end


c     ==== copy one matrix to another element-by-element ====
      subroutine cpmat(g_in,g_out)
      real*8   g_in(4,4),g_out(4,4)
      do j=1,4
         do i=1,4
            g_out(i,j)=g_in(i,j)
         enddo
      enddo
      return
      end


c     --- This calls lapack to calculate an inverse, but
c     is not really needed, since the inverse is the transverse
c     in the case of an orthonormal basis, so to reduce unnecessary
c     compilation dependencies, this is commented out.

C      subroutine getinv(g_in,g_out)
C      include 'eigcom.f'
C
CC     INPUTS: 
C      real*8   g_in(4,4)
C
CC     OUTPUTS:
C      real*8   g_out(4,4)
C
Cc     INTERNAL:
C      real*8 gen_lu(4,4) ! working copy of matrix, e.g. for LU factorisation
C      parameter(lwork=256) 
C      real*8 work(lwork) ! array for working space
C
C      integer  ipiv(4) ! pivot indices, indicating permutations
C      
C
C      do j=1,4
C         do i=1,4
C            gen_lu(i,j)= g_in(i,j)
C         enddo
C      enddo
C
Cc      if(q_verb)then
Cc         print*,'Before LU:'
Cc         call printmat(gen_lu)
Cc      endif
C
Cc     --- do the LU factorization ---
C      call dgetrf(4,4, gen_lu, 4, ipiv, info)
C      if(info.ne.0)then
C         print*,'LU factorisation with dgetrf failed. '
C         call printmat(g_in)
C      else
Cc         if(q_verb)then
Cc            print*,'LU factorisation is'
Cc            call printmat(gen_lu)
Cc         endif
C         call dgetri(4, gen_lu, 4, ipiv, work, lwork, info)
C         if(info.ne.0)then
C            print*,'Error in inverting matrix with  dgetri.'
C            print*,'Edit eigmain.f, recompile, rerun for more info.'
C         endif
C      endif
C
Cc      if(q_verb)then
Cc         print*,'After degtri :'
Cc         call printmat(gen_lu)
Cc      endif
C
C      do j=1,4
C         do i=1,4
C            g_out(i,j)= gen_lu(i,j)
C         enddo
C      enddo
C
C      return
C      end


c     === set a 4x4 real*8  matrix to all zero values ===
      subroutine zeromat(gen)
C     INPUT modified to OUTPUT
      real*8   gen(4,4)
      
      do j=1,4
         do i=1,4
            gen(i,j)=0.0d0
         enddo
      enddo
      return
      end

            

      subroutine transp(g_in,g_out)
      include 'eigcom.f'

C     INPUTS: 
      real*8   g_in(4,4)

C     OUTPUTS:
      real*8   g_out(4,4)

      do i=1,4
         do j=1,4
            g_out(i,j)=g_in(j,i)
         enddo
      enddo
      return
      end


c     ---- debugging test ----
      subroutine test_1(gen)
      include 'eigcom.f'

C     INPUTS:
      real*8     gen(4,4)
C     INTERNAL:
      real*8     uguess(4)

c     ---- test getchi routine with vectors known to be in eigenplanes ---
c     ---- start with an arbitrary vector ----
      do i=1,4
         uguess(i)=0.0d0
      enddo
      x=randt()
      uguess(1)=randt()         !1.0d0/sqrt(2.0d0)
      uguess(2)=sqrt(1.0d0 - uguess(1)*uguess(1))

      if(debug)print*,'BB- uguess=',(uguess(i),i=1,4)

      call getchi(uguess, gen, chi, ang_u)
      print*,'uguess=',(uguess(i),i=1,4)
      print*,'chi,ang_u=',chi,ang_u

      do i=1,4
         uguess(i)=0.0d0
      enddo
      uguess(2)=1.0d0
      call getchi(uguess, gen, chi, ang_u)
      print*,'uguess=',(uguess(i),i=1,4)
      print*,'chi,ang_u=',chi,ang_u

      do i=1,4
         uguess(i)=0.0d0
      enddo
      uguess(3)=1.0d0
      call getchi(uguess, gen, chi, ang_u)
      print*,'uguess=',(uguess(i),i=1,4)
      print*,'chi,ang_u=',chi,ang_u

      do i=1,4
         uguess(i)=0.0d0
      enddo
      uguess(4)=1.0d0
      call getchi(uguess, gen, chi, ang_u)
      print*,'uguess=',(uguess(i),i=1,4)
      print*,'chi,ang_u=',chi,ang_u

      return
      end
      

c     ==== generate an isometry in a random basis ====
      subroutine getiso(genbas,thet_s,phi_s)
      include 'eigcom.f'

C     OUTPUTS: 
      real*8  genbas(4,4)  ! "input" matrix
      real*8  thet_s,phi_s ! rotation angles

C     INTERNAL:
      real*8  gen(4,4),gtemp(4,4) ! working matrices
      real*8  rbasis(4,4)       ! random basis
      real*8  rbinv(4,4) ! inverse of rbasis


      character*1  ch

      logical*4    qrotb  ! rotate the random basis several times?
      data  qrotb /.false./ ! .true. / !

      data ch /'n'/


      if(q_verb)print*,'Generate a random isometry of S^3 in R^4',
     _     ' expressed in a randomly oriented orthonormal basis'


      if(q_verb)print*,'Generate a random isometry of S^3 in R^4...'
c     ---- generate the matrix - axis choices and angles are random ---
      call r_isom(gen,thet_s,phi_s)  

c     -- the isometry is now   gen  --

c     --- determinant calculation is only a check - it is not needed ---
c      if(debug)then
c         det= det4x4(gen)       ! find the determinant
c         if(q_verb)print*,"Its determinant is: ", det
c      endif

      if(debug) call test_1(gen)


c     -- now find a basis --
      if(q_verb)print*,'Generate a random orthonormal basis...'
c     -- first rotate the identity basis in two complementary planes
      call r_isom(rbasis,th1,th2) 

c     -- optionally, successively rotate in several arbitrarily chosen planes --
c     
c     The point of these successive rotations is to make sure that the
c     basis is sufficiently "random" and unaligned with the xyzw axes. 
c     However, numerical error accumulates and the final matrix
c     rbasis seems to be not quite orthonormal. The result is that
c     the generator expressed in this not-quite-orthonormal basis,
c     rbinv * gen * rbasis 
c     does not have quite the same rotation angles thet_s, phi_s  as
c     gen
c     This makes convergence and precision testing difficult.
c
      if(qrotb)then
         theta=randt()*pi
         call ranrot(1,2,theta,rbasis,gtemp)
         call cpmat(gtemp,rbasis)
         theta=randt()*pi
         call ranrot(1,3,theta,rbasis,gtemp)
         call cpmat(gtemp,rbasis)
         theta=randt()*pi
         call ranrot(1,4,theta,rbasis,gtemp)
         call cpmat(gtemp,rbasis)
         theta=randt()*pi
         call ranrot(2,3,theta,rbasis,gtemp)
         call cpmat(gtemp,rbasis)
c     
c     If you wish to use this (conservative) randomising of the basis,
c     then you should probably "re-orthonormalise" rbasis at this point.
c     Only minor nudges should be necessary, but if you don't do them,
c     then you'll probably have numerical precision problems.

      endif

      if(q_verb)then
         print*,'The arbitrarily chosen basis is: '
         call printmat(rbasis)
      endif

c     --- inverse of  rbasis  using  dtrsv from blas library ----
c      call getinv(rbasis,rbinv)
      call transp(rbasis,rbinv) ! transpose is inverse for othonormal matrix

      if(q_verb)then
         print*,"Its inverse is: "
         call printmat(rbinv)
         print*,'and  rbasis * rbinv ='
         call dgemm(ch,ch, 4,4,4, 1.0d0, rbasis, 4, rbinv, 4, 
     _        0.0d0, gtemp, 4)
         call printmat(gtemp)
      endif
      
c      if(debug)then
c         det= det4x4(rbasis)    ! find the determinant
c         if(q_verb)print*,"Its determinant is: ", det
c      endif

c     ---- matrix  gen  expressed in the basis   rbasis  is  
c     inv(rbasis) * gen * rbasis 
c     i.e.   rbinv * gen * rbasis 
c     The two steps to calculate this are:

      call dgemm(ch,ch, 4,4,4, 1.0d0, rbinv, 4, gen, 4, 
     _        0.0d0, gtemp, 4)

      call dgemm(ch,ch, 4,4,4, 1.0d0, gtemp, 4, rbasis, 4, 
     _        0.0d0, genbas, 4)

c      print*,'!! TEST ONLY !! overriding genbas'
c      call cpmat(gen,genbas)


      print*,'A random isometry   genbas  has been calculated '
      print*,'with the two rotation angles'
      print*,'thet_s, phi_s=', thet_s,phi_s 

      print*,' '
      print*,'A randomly aligned orthonormal basis was chosen.'
      print*,'It most probably does not align with'
      print*,'either eigenplane. '
      print*,'In this basis, this isometry is represented as:'
      call printmat(genbas)
      print*,' '
      if(q_verb)then
         print*,'From this matrix alone, it should be possible,'
         print*,"by the algorithm below, to find",
     _        " the 'secret' rotation"
         print*,'angles thet_s and phi_s'
         print*,' ' 
      endif
      

      return
      end 


c     ===
      subroutine getang(genbas, u0, thet_s, vec)
      include 'eigcom.f'

C     INPUTS:
      real*8  genbas(4,4)  ! "input" matrix
      real*8  u0(4)  ! initial guess

C     OUTPUTS: 
      real*8  thet_s            ! rotation angle
      real*8  vec(4)            ! a vector in an eigenplane

C     INTERNAL:
      real*8 uguess(4), v(4), vguess(4)
      real*8  gtemp(4,4)  ! working matrix

c      character*1  ch
c      data  ch /'n'/

      if(q_verb)print*,'Looping through rotations ',
     _     'in the 1-2 and 1-3 planes...'

c     ---- set initial rotation limits to +- pi/2 ----
c     -- of course, rotation is not a commutative operation, but this shouldn't
c     -- stop convergence of the algorithm 

      dth=pi*(1.0d0+toldef) /2.0d0

c      if(debug)print*,'CC- uguess=',(uguess(i),i=1,4)
c      if(debug)uguess(1)=randt()         !1.0d0/sqrt(2.0d0)
c      if(debug)print*,'CC2- uguess=',(uguess(i),i=1,4)

c     --- copy initial guess ---
      do i=1,4
         uguess(i)=u0(i)
      enddo


c     ---- how good is this first wild guess? ----
      call getchi(uguess, genbas, chi, ang_u)
      if(q_verb)print*,'uguess=',(uguess(i),i=1,4)
      if(q_verb)print*,'chi,ang_u=',chi,ang_u


c     -- test the four adjacent values and use the best as the new best
c     estimate --
      safetyfactor = 5.0d0      ! should not need to be so big
      imax = int(log10(precision)/log10(f_theta) * safetyfactor)+1
      if(q_verb)print*,'imax,precision=',imax,precision
      ichi=1
      chiold=chi
      do while(ichi.le.imax.and.
     _     dth.gt.precision)
c     _     chi.gt.precision)
         
         if(q_verb)then
            print*,'--------------- ' 
            print*,'ichi=',ichi
         endif

c     ---- copy best guess from last loop to tmp working guess:  vguess  --
         do i=1,4
            vguess(i)=uguess(i)
         enddo
      

c     --- check in each of the six directions 
         do k=1,6
            if(k.le.2)then
               i1=1
               i2=2 ! do a 1-2 rotation
            elseif(k.le.4)then
               i1=1
               i2=3 ! do a 1-3 rotation
            else
               i1=2
               i2=3 ! do a 2-3 rotation
            endif
            if(k.eq.2.or.k.eq.4.or.k.eq.6)then
               th1= -dth        ! negative angle
            else
               th1= dth
            endif

            if(q_verb)print*,' '
            if(q_verb)print*,'i1,i2,th1=',i1,i2,th1
            call idrot(i1,i2,th1,gtemp) ! rotation in i1-i2 plane by angle th1 

            if(q_verb)print*,'uguess=',(uguess(i),i=1,4),' gtemp ='
            if(q_verb)call printmat(gtemp)


c     -- apply rotation to best guess  uguess --
            call dgemv('n', 4,4, 1.0d0, gtemp, 4, uguess,
     _           1, 0.0d0, v, 1)

            if(q_verb.and.debug)then
               write(6,'(a,$)')'will try  v =         '
               write(6,'(4e10.2)')(v(i),i=1,4)
            endif

c     -- how good is this shifted guess? --
            call getchi(v, genbas, chi1, a)

            if(chi1.lt.chi)then ! if better, then store it temporarily!
               chi=chi1
               do i=1,4
                  vguess(i)=v(i)
               enddo
               ang_u = a
               if(q_verb.and.debug)then
                  print*,'ichi, chi=',ichi,chi,' and vguess is: '
                  write(6,'(4e10.2)')(vguess(i),i=1,4)
               endif
            endif
         enddo                  !          do k=1,4
         
c     --- prepare for next loop
c     -- if no better guess was found, this may simply be the previous
c     best guess --
         do i=1,4
            uguess(i)=vguess(i)
         enddo
      
c         if(chi.lt.0.99d0*chiold)then
         dth= dth * f_theta     ! improve precision for next loop
c         else
c            dth= 0.95 *dth ! only improve precision slowly if it's not improving
c         endif

         ichi=ichi+1            ! loop safety check
         chiold=chi
      enddo                     !      do while(i.le.imax.and.chi.gt.precision)

      if(q_verb)then
         print*,'ichi=',ichi
         print*,'One vector of one of the eigenplanes seems to be:'
         write(6,'(4f8.2)')(uguess(i),i=1,4)
         print*,'with angle ', ang_u
         print*,' '
      endif

      do i=1,4
         vec(i)=uguess(i)
      enddo

      thet_s= ang_u

      return
      end


c     ==== starting from one vector soln(.,1), find the other three ===
      subroutine getfour(genbas,soln,thet_s,phi_s)
      include 'eigcom.f'
      
C     INPUTS:
      real*8   genbas(4,4)      ! the isometry

C     MODIFIED:
      real*8   soln(4,4)        ! the first vector (.,1) should be INPUT

C     OUTPUTS:
      real*8   thet_s,phi_s

C     INTERNAL:
      real*8   u(4) ! working vector
      real*8   ortho(4,3)          ! three orthonormal vectors for trying out
      real*8   size(3)          ! size after subtracting first 2 components

c     --- apply isometry to first vector
      call dgemv('n', 4,4, 1.0d0, genbas, 4, soln(1,1),
     _     1, 0.0d0, u, 1)
      
      costheta= ddot(4, soln(1,1), 1, u, 1)
      thet_s= acos(costheta)

      sintheta=sqrt( max(1.0d0-costheta*costheta, toldef) )
      sinv= 1.0d0/ sintheta  ! avoid dividing more often than needed 

c     --- get second vector by subtracting its projection on the first
c     and normalising ---

      do i=1,4
         soln(i,2)= (u(i) - costheta* soln(i,1)) *sinv
      enddo


      do itry=1,3
c     --- at least one of these 3 vectors must be out of the 1st eigenplane ---
         do i=1,4
            u(i)=0.0d0
         enddo
         u(itry)=1.0d0

         do i=1,4
            ortho(i,itry)=u(i)
         enddo

c     --- subtract the components in the 1st, 2nd directions ---
         do isoln=1,2
            costheta= ddot(4, soln(1,isoln), 1, u, 1)
            sintheta=sqrt( max(1.0d0-costheta*costheta, toldef) )
            sinv= 1.0d0/ sintheta ! avoid dividing more often than needed 
            
c     --- get second vector by subtracting its projection on the first
c     and normalising ---
            do i=1,4
               ortho(i,itry)= (ortho(i,itry) - 
     _              costheta* soln(i,isoln)) *sinv
            enddo
         enddo

c     -- square value of modulus is sufficient for this test --
         size(itry)= ddot(4, ortho(1,itry),1, ortho(1,itry),1) 

      enddo     !   do itry=1,3

      i=1
      s=size(1)
      do itry=2,3
         if(size(itry).gt.s)then
            i= itry
            s= size(itry)
         endif
      enddo
      itrybest= i

      ssqrt= 1.0d0/sqrt(size(itrybest))
      do i=1,4
         soln(i,3)= ortho(i,itrybest) * ssqrt
      enddo

c     --- apply isometry to third vector
      call dgemv('n', 4,4, 1.0d0, genbas, 4, soln(1,3),
     _     1, 0.0d0, u, 1)
      
      costheta= ddot(4, soln(1,3), 1, u, 1)
      phi_s= acos(costheta)

      sintheta=sqrt( max(1.0d0-costheta*costheta, toldef) )
      sinv= 1.0d0/ sintheta  ! avoid dividing more often than needed 

c     --- get second vector by subtracting its projection on the first
c     and normalising ---

      do i=1,4
         soln(i,4)= (u(i) - costheta* soln(i,3)) *sinv
      enddo

      return
      end


