C   eigenplane - find basis vectors for eigenplanes of an isometry of S^3 represented as a 4x4 matrix
C
C   Copyright (C) 2004 Boud Roukema
C
C   This program is free software; you can redistribute it and/or modify
C   it under the terms of the GNU General Public License as published by
C   the Free Software Foundation; either version 2, or (at your option)
C   any later version.
C
C   This program is distributed in the hope that it will be useful,
C   but WITHOUT ANY WARRANTY; without even the implied warranty of
C   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C   GNU General Public License for more details.
C
C   You should have received a copy of the GNU General Public License
C   along with this program; if not, write to the Free Software Foundation,
C   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  


C     convert the command line parameters to a fortran common block and start
C     the main eigenplane subroutine

      subroutine mainsub(oname,ioname, iname,iiname,
     _     i_int,i_ran,i_v,i_w,chseed,ichseed,
     _     chprec,ichprec)

c      implicit real*8 (a-h,p-z)
      include 'eigcom.f'


      character*(*)  oname, chseed, chprec, iname
      integer        ioname, ichseed, ichprec, iiname
      integer        i_int,i_ran,i_v,i_w

c     INTERNAL:
      character*1    ch
      character*256  chlong

c     ==== command line parameters ====

c     -- Warning - these are not defaults - defaults should be set
c     in the calling C program --
      q_int=.false.
      q_random=.false.
      q_verb=.false.
      q_warn=.false.
      q_output=.false.
      q_input=.false.

      if(i_int.gt.0)q_int=.true.
      if(i_ran.gt.0)q_random=.true.
      if(i_v.gt.0)q_verb=.true.
      if(i_w.gt.0)q_warn=.true.

      chlong=oname(1:min(ioname,256))

      if(debug)print*,'ioname,chlong(1:ioname)=',
     _     ioname,chlong(1:ioname)

c      if(oname(1:ioname).ne.'stdout')q_output=.true.
      if(chlong(1:ioname).ne.'stdout')q_output=.true.
      if(debug)print*,'q_output=',q_output



      chlong=iname(1:min(iiname,256))

      if(debug)print*,'iiname,chlong(1:iiname)=',
     _     iiname,chlong(1:iiname)

      if(chlong(1:iiname).ne.'stdin')q_input=.true.
      if(debug)print*,'q_input=',q_input



      if(chseed(1:ichseed).eq.'none')then
         iseed = 123456           ! default
      else
         if(q_verb)print*,'Will try reading in random seed'
         print*,'chseed=  ',chseed(1:ichseed)
         read(chseed(1:ichseed),*)iseed
      endif
      if(q_verb)print*,'random seed  iseed=',iseed

      if(chprec(1:ichprec).eq.'none')then
         precision = 1d-6          ! default
      else
         if(q_verb)print*,'Will try reading in precision'
         print*,'chprec=  ',chprec(1:ichprec)
         read(chprec(1:ichprec),*)precision
      endif
      if(q_verb)print*,'precision for finding angles=',precision

      if(q_int)then
         print*,'--interactive: This will be run interactively.'
         if(.not.(q_verb))then
            print*,'Consider using the -v or --verbose option '
            print*,'in order to see more diagnostic information.'
         endif
         write(6,'(a,$)')'Type any key and enter to continue: '
         read(5,*)ch
      endif

      if(.not.(q_verb))then
         continue
c         print*,'Very little output (except this line) will be printed.'
      else
         print*,'--verbose: You chose the verbose option,'
         print*,'so a huge amount'
         print*," of diagnostic info may be printed. Don't"
         print*,"say you weren't warned! ;)" 
      endif

      if(q_output)then
         iofile=16
      else
         iofile=6 ! stdout
      endif
      if(q_input)then
         iifile=15
      else
         iifile=5 ! stdin
      endif


      chlong=oname(1:ioname)


      if(q_verb)print*,'--output: Will try writing some output to: ',
     _     chlong(1:ioname),' ...'

      if(q_output)open(iofile,file=chlong,
     _     access='sequential',status='unknown')

      if(debug)write(iofile,'(a)')
     _     'eigenplane output file - first line' ! may be stdout

      if(q_output)close(iofile)
      if(q_verb)print*,'Finished writing some output to: ',
     _     chlong(1:ioname)

      if(q_warn)print*,'--warning-disable: WARNING:',
     _     'you will get no more warnings since you have switched', 
     _     'them off; however, lethal errors will still crash the',
     _     'program.'

c     ==== copy command line parameters to common block ====

      if(ioname.gt.256)then
         onamc=chlong(1:256)
         ionamc=256
         if(q_warn)then
            print*,'WARNING: output filename will be truncated: '
            print*,onamc
         endif
      else
         onamc=chlong(1:ioname)
         ionamc=ioname
      endif

      if(debug)print*,'B: ioname,onamc(1:ionamc)=',
     _     ioname,onamc(1:ionamc)
      if(debug)print*,'q_output=',q_output


      chlong=iname(1:iiname)
      if(iiname.gt.256)then
         inamc=chlong(1:256)
         iinamc=256
         if(q_warn)then
            print*,'WARNING: input filename will be truncated: '
            print*,inamc
         endif
      else
         inamc=chlong(1:iiname)
         iinamc=iiname
      endif

      if(debug)then
         print*,'inamc(1:iinamc),q_input=',inamc(1:iinamc),q_input
      endif

c      print*,'mainsub: toldef=',toldef

c     ==== call main eigenplane subroutine ====
c
      call eigmain
c

      return
      end

